@@ -100,7 +100,7 @@ set_unknown_glyph (PangoGlyphString *glyphs,
 #include "hangul-defs.h"
 
 typedef void (* RenderSyllableFunc) (PangoFont *font, PangoXSubfont subfont,
-				     gunichar2 *text, int length,
+				     gunichar *text, int length,
 				     PangoGlyphString *glyphs,
 				     int *n_glyphs, int cluster_offset);
 
@@ -139,33 +139,19 @@ typedef void (* RenderSyllableFunc) (PangoFont *font, PangoXSubfont subfont,
   if (n_cho <= 1 && n_jung <= 1)					      \
     {									      \
       int composed = 0;							      \
-      gunichar2 l, v, t;						      \
+      gunichar l, v, t = 0;						      \
 									      \
-      if (n_cho > 0)							      \
-	l = text[0];							      \
-      else								      \
-	l = LFILL;							      \
-									      \
-      if (n_jung > 0)							      \
-	v = text[n_cho];						      \
-      else								      \
-	v = VFILL;							      \
-									      \
-      if (n_jong > 0)							      \
-	t = text[n_cho + n_jung];					      \
-      else								      \
-	t = TFILL;							      \
+      l = (n_cho > 0) ? text[0] : LFILL;				      \
+      v = (n_jung > 0) ? text[n_cho] : VFILL;				      \
+      t = (n_jong > 0) ? text[n_cho + n_jung] : 0;			      \
 									      \
       if ((__choseong_johabfont_base[l - LBASE] != 0 || l == LFILL) &&	      \
 	  (__jungseong_johabfont_base[v - VBASE] != 0 || v == VFILL) &&	      \
-	  (__jongseong_johabfont_base[t - TBASE] != 0 || t == TFILL))	      \
+	  (!t || (__jongseong_johabfont_base[t - TBASE] != 0)))		      \
 	composed = n_cho + n_jung + ((n_jong > 0) ? 1 : 0);		      \
       else if ((__choseong_johabfont_base[l - LBASE] != 0 || l == LFILL) &&   \
 	       (__jungseong_johabfont_base[v - VBASE] != 0 || v == VFILL))    \
-	{								      \
-	  composed = n_cho + n_jung;					      \
-	  t = TFILL;							      \
-	}								      \
+	composed = n_cho + n_jung;					      \
 									      \
       /* COMPOSABLE */							      \
       if (composed)							      \
@@ -173,7 +159,7 @@ typedef void (* RenderSyllableFunc) (PangoFont *font, PangoXSubfont subfont,
 	  if (l != LFILL)						      \
 	    {								      \
 	      gindex = __choseong_johabfont_base[l - LBASE];		      \
-	      if (t == TFILL)						      \
+	      if (!t)							      \
 		{							      \
 		  if (v == VFILL)					      \
 		    gindex += 1;					      \
@@ -199,11 +185,12 @@ typedef void (* RenderSyllableFunc) (PangoFont *font, PangoXSubfont subfont,
 	      switch (__johabfont_jungseong_kind[v - VBASE])		      \
 		{							      \
 		case 3:							      \
-		  gindex += __johabfont_jongseong_kind[t - TBASE];	      \
+		  if (t)						      \
+		    gindex += __johabfont_jongseong_kind[t - TBASE];	      \
 		  break;						      \
 		case 4:							      \
 		  gindex += ((l == 0x1100 || l == 0x110f) ? 0 : 1) +	      \
-		    ((t != TFILL) ? 2 : 0);				      \
+		    (t ? 2 : 0);					      \
 		  break;						      \
 		}							      \
   									      \
@@ -213,7 +200,7 @@ typedef void (* RenderSyllableFunc) (PangoFont *font, PangoXSubfont subfont,
 	      (*n_glyphs)++;						      \
 	    }								      \
 									      \
-	  if (t != TFILL)						      \
+	  if (t)							      \
 	    {								      \
 	      gindex = __jongseong_johabfont_base[t - TBASE] +		      \
 		__jongseong_map[v - VBASE];				      \
@@ -223,7 +210,7 @@ typedef void (* RenderSyllableFunc) (PangoFont *font, PangoXSubfont subfont,
 	      (*n_glyphs)++;						      \
 	    }								      \
 									      \
-	  if (v == VFILL && t == TFILL) /* dummy for no zero width */	      \
+	  if (v == VFILL && !t) /* dummy for no zero width */		      \
 	    {								      \
 	      pango_glyph_string_set_size (glyphs, *n_glyphs + 1);	      \
 	      set_glyph (glyphs, *n_glyphs, font, subfont, JOHAB_FILLER);     \
@@ -238,10 +225,11 @@ typedef void (* RenderSyllableFunc) (PangoFont *font, PangoXSubfont subfont,
 
 static void
 render_syllable_with_johabs (PangoFont *font, PangoXSubfont subfont,
-			     gunichar2 *text, int length,
+			     gunichar *text, int length,
 			     PangoGlyphString *glyphs,
 			     int *n_glyphs, int cluster_offset)
 {
+  int n_prev_glyphs = *n_glyphs;
   JOHAB_COMMON
 
   /* Render as uncomposed forms as a fallback.  */
@@ -249,7 +237,7 @@ render_syllable_with_johabs (PangoFont *font, PangoXSubfont subfont,
     {
       int j;
 
-      if (text[i] == LFILL || text[i] == VFILL || text[i] == TFILL)
+      if (text[i] == LFILL || text[i] == VFILL)
 	continue;
       /*
        * Uses KSC5601 symbol glyphs which johabS-1 has; they're
@@ -275,15 +263,24 @@ render_syllable_with_johabs (PangoFont *font, PangoXSubfont subfont,
       if (j == 0)
 	set_unknown_glyph (glyphs, n_glyphs, font, text[i], cluster_offset);
     }
-
+  /* No glyph rendered yet; it means the empty syllable (ex. LFILL+VFILL) */
+  if (n_prev_glyphs == *n_glyphs)
+    {
+      gindex = JOHAB_FILLER;
+      pango_glyph_string_set_size (glyphs, *n_glyphs + 1);
+      set_glyph (glyphs, *n_glyphs, font, subfont, gindex);
+      glyphs->log_clusters[*n_glyphs] = cluster_offset;
+      (*n_glyphs)++;
+    }
 }
 
 static void
 render_syllable_with_johab (PangoFont *font, PangoXSubfont subfont,
-			    gunichar2 *text, int length,
+			    gunichar *text, int length,
 			    PangoGlyphString *glyphs,
 			    int *n_glyphs, int cluster_offset)
 {
+  int n_prev_glyphs = *n_glyphs;
   JOHAB_COMMON
 
   /* Render as uncomposed forms as a fallback.  */
@@ -291,7 +288,7 @@ render_syllable_with_johab (PangoFont *font, PangoXSubfont subfont,
     {
       int j;
 
-      if (text[i] == LFILL || text[i] == VFILL || text[i] == TFILL)
+      if (text[i] == LFILL || text[i] == VFILL)
 	continue;
       for (j = 0; (j < 3) && (__jamo_to_johabfont[text[i]-LBASE][j] != 0); j++)
 	{
@@ -311,16 +308,24 @@ render_syllable_with_johab (PangoFont *font, PangoXSubfont subfont,
       if (j == 0)
 	set_unknown_glyph (glyphs, n_glyphs, font, text[i], cluster_offset);
     }
-  if (*n_glyphs == 0)
-    set_unknown_glyph (glyphs, n_glyphs, font, 0, cluster_offset);
+  /* No glyph rendered yet; it means the empty syllable (ex. LFILL+VFILL) */
+  if (n_prev_glyphs == *n_glyphs)
+    {
+      gindex = JOHAB_FILLER;
+      pango_glyph_string_set_size (glyphs, *n_glyphs + 1);
+      set_glyph (glyphs, *n_glyphs, font, subfont, gindex);
+      glyphs->log_clusters[*n_glyphs] = cluster_offset;
+      (*n_glyphs)++;
+    }
 }
 
 static void
-render_syllable_with_ksx1005 (PangoFont *font, PangoXSubfont subfont,
-			       gunichar2 *text, int length,
-			       PangoGlyphString *glyphs,
-			       int *n_glyphs, int cluster_offset)
+render_syllable_with_ksx1001johab (PangoFont *font, PangoXSubfont subfont,
+				   gunichar *text, int length,
+				   PangoGlyphString *glyphs,
+				   int *n_glyphs, int cluster_offset)
 {
+  int n_prev_glyphs = *n_glyphs;
   guint16 gindex;
   int i, composed;
 
@@ -359,7 +364,7 @@ render_syllable_with_ksx1005 (PangoFont *font, PangoXSubfont subfont,
     {
       int j;
 
-      if (text[i] == LFILL || text[i] == VFILL || text[i] == TFILL)
+      if (text[i] == LFILL || text[i] == VFILL)
 	continue;
 
       gindex = text[i];
@@ -369,7 +374,7 @@ render_syllable_with_ksx1005 (PangoFont *font, PangoXSubfont subfont,
           if (index >= 0x2400 && index < 0x2500)
             index = 0xda80 + index % 256;
           else /* 0x2300 - 0x2400 */
-            index = 0xda10 + index % 256;
+            index = 0xda10 + index % 256 + (index > 0x236e ? 0x12 : 0);
 	  pango_glyph_string_set_size (glyphs, *n_glyphs + 1);
 	  set_glyph (glyphs, *n_glyphs, font, subfont, index);
 	  glyphs->log_clusters[*n_glyphs] = cluster_offset;
@@ -378,14 +383,24 @@ render_syllable_with_ksx1005 (PangoFont *font, PangoXSubfont subfont,
       if (j == 0)
 	set_unknown_glyph (glyphs, n_glyphs, font, gindex, cluster_offset);
     }
+  /* No glyph rendered yet; it means the empty syllable (ex. LFILL+VFILL) */
+  if (n_prev_glyphs == *n_glyphs)
+    {
+      gindex = 0xd931;
+      pango_glyph_string_set_size (glyphs, *n_glyphs + 1);
+      set_glyph (glyphs, *n_glyphs, font, subfont, gindex);
+      glyphs->log_clusters[*n_glyphs] = cluster_offset;
+      (*n_glyphs)++;
+    }
 }
 
 static void
 render_syllable_with_iso10646 (PangoFont *font, PangoXSubfont subfont,
-			       gunichar2 *text, int length,
+			       gunichar *text, int length,
 			       PangoGlyphString *glyphs,
 			       int *n_glyphs, int cluster_offset)
 {
+  int n_prev_glyphs = *n_glyphs;
   guint16 gindex;
   int i, composed;
 
@@ -413,16 +428,28 @@ render_syllable_with_iso10646 (PangoFont *font, PangoXSubfont subfont,
   /* Render as uncomposed forms as a fallback.  */
   for (i = 0; i < length; i++)
     {
+      if (text[i] == LFILL || text[i] == VFILL)
+	continue;
+
       gindex = text[i];
       pango_glyph_string_set_size (glyphs, *n_glyphs + 1);
       set_glyph (glyphs, *n_glyphs, font, subfont, gindex);
       glyphs->log_clusters[*n_glyphs] = cluster_offset;
       (*n_glyphs)++;
     }
+  /* No glyph rendered yet; it means the empty syllable (ex. LFILL+VFILL) */
+  if (n_prev_glyphs == *n_glyphs)
+    {
+      gindex = 0x3164;
+      pango_glyph_string_set_size (glyphs, *n_glyphs + 1);
+      set_glyph (glyphs, *n_glyphs, font, subfont, gindex);
+      glyphs->log_clusters[*n_glyphs] = cluster_offset;
+      (*n_glyphs)++;
+    }
 }
 
 static int
-find_ksc5601 (gunichar2 ucs)
+find_ksc5601 (gunichar ucs)
 {
   int l = 0;
   int u = KSC5601_HANGUL - 1;
@@ -450,9 +477,9 @@ find_ksc5601 (gunichar2 ucs)
 }
 
 static gboolean
-try_decompose_jongseong(gunichar2 jong, gunichar2 *jong1, gunichar2 *jong2)
+try_decompose_jongseong(gunichar jong, gunichar *jong1, gunichar *jong2)
 {
-  static gunichar2 table[][2] = {
+  static gunichar table[][2] = {
     {0, 0},
     /* U+11A8- */
     {0, 0}, {0, 0}, {0x11A8, 0x11BA}, {0, 0},
@@ -496,12 +523,13 @@ try_decompose_jongseong(gunichar2 jong, gunichar2 *jong1, gunichar2 *jong2)
 
 static void
 render_syllable_with_ksc5601 (PangoFont *font, PangoXSubfont subfont,
-			      gunichar2 *text, int length,
+			      gunichar *text, int length,
 			      PangoGlyphString *glyphs,
 			      int *n_glyphs, int cluster_offset)
 {
+  int n_prev_glyphs = *n_glyphs;
   int gindex = 0, composed = 0;
-  gunichar2 jong1, jong2 = 0;
+  gunichar jong1, jong2 = 0;
   int i;
 
   if (length >= 3 && IS_L_S(text[0]) && IS_V_S(text[1]) && IS_T_S(text[2]))
@@ -567,6 +595,15 @@ render_syllable_with_ksc5601 (PangoFont *font, PangoXSubfont subfont,
       if (j == 0)
 	set_unknown_glyph (glyphs, n_glyphs, font, gindex, cluster_offset);
     }
+  /* No glyph rendered yet; it means the empty syllable (ex. LFILL+VFILL) */
+  if (n_prev_glyphs == *n_glyphs)
+    {
+      gindex = 0x2121;
+      pango_glyph_string_set_size (glyphs, *n_glyphs + 1);
+      set_glyph (glyphs, *n_glyphs, font, subfont, gindex);
+      glyphs->log_clusters[*n_glyphs] = cluster_offset;
+      (*n_glyphs)++;
+    }
 }
 
 static gboolean
@@ -627,7 +664,7 @@ find_subfont (PangoFont *font, char **charsets, int n_charsets,
       else if (strcmp (charsets[subfont_charsets[i]], "ksc5601.1992-3") == 0)
 	{
 	      *subfont = subfonts[i];
-	      *render_func = render_syllable_with_ksx1005;
+	      *render_func = render_syllable_with_ksx1001johab;
 	      break;
 	}
       else if (strcmp (charsets[subfont_charsets[i]], "ksc5601.1987-0") == 0)
@@ -668,15 +705,14 @@ hangul_engine_shape (PangoFont        *font,
   PangoXSubfont subfont;
   RenderSyllableFunc render_func = NULL;
 
-  const char *ptr;
-  const char *next;
-  int i, n_chars;
-  gunichar2 jamos_static[4];
-  guint max_jamos = G_N_ELEMENTS (jamos_static);
-  gunichar2 *jamos = jamos_static;
-  int n_jamos = 0;
+  int n_chars, n_glyphs;
+  int i;
+  const char *p, *start;
 
-  int n_glyphs = 0, cluster_offset = 0;
+  gunichar jamos_static[8];
+  guint max_jamos = G_N_ELEMENTS (jamos_static);
+  gunichar *jamos = jamos_static;
+  int n_jamos = 0;
 
   g_return_if_fail (font != NULL);
   g_return_if_fail (text != NULL);
@@ -697,95 +733,62 @@ hangul_engine_shape (PangoFont        *font,
 	}
 
   n_chars = g_utf8_strlen (text, length);
-  ptr = text;
+  n_glyphs = 0;
+  start = p = text;
+  n_jamos = 0;
   
   for (i = 0; i < n_chars; i++)
     {
-      gunichar wc4;
-      gunichar2 wcs[4], wc;
-      int n_code = 0;
+      gunichar prev = jamos[n_jamos - 1];
+      gunichar wc;
 
-      wc4 = g_utf8_get_char (ptr);
-      next = g_utf8_next_char (ptr);
-      
-      if (wc4 >= SBASE && wc4 < (SBASE + SCOUNT))
+      wc = g_utf8_get_char (p);
+
+      /* Check syllable boundaries. */
+      if (n_jamos &&
+	  ((!IS_L (prev) && IS_S (wc)) ||
+	   (IS_T (prev) && IS_L (wc)) ||
+	   (IS_V (prev) && IS_L (wc)) ||
+	   (IS_T (prev) && IS_V (wc))))
 	{
-	  /* decompose the syllable.  */
-	  gint16 sindex;
+	  /* Draw a syllable. */
 
-	  sindex = wc4 - SBASE;
-	  wcs[0] = LBASE + (sindex / NCOUNT);
-	  wcs[1] = VBASE + ((sindex % NCOUNT) / TCOUNT);
-	  wcs[2] = TBASE + (sindex % TCOUNT);
-	  n_code = 3;
-
-	  if (n_jamos > 0)
-	    {
-	      (*render_func) (font, subfont, jamos, n_jamos,
-			      glyphs, &n_glyphs, cluster_offset);
-	      cluster_offset = next - text;
-	      n_jamos = 0;
-	    }
-
-	  /* Draw a syllable.  */
-	  (*render_func) (font, subfont, wcs, 3,
-			  glyphs, &n_glyphs, cluster_offset);
-	  cluster_offset = next - text;
-	  /* Clear.  */
+	  (*render_func) (font, subfont, jamos, n_jamos,
+			  glyphs, &n_glyphs, start - text);
+	  n_jamos = 0;
+	  start = p;
 	}
-      else if (wc4 >= 0x1100 && wc4 <= 0x11ff)
-	{
-	  wc = (gunichar2) wc4;
 
-	  if (n_jamos == 0)
+      if (n_jamos == max_jamos)
+	{
+	  max_jamos += 3;	/* at most 3 for each syllable code (L+V+T) */
+	  if (jamos == jamos_static)
 	    {
-	      jamos[n_jamos++] = wc;
+	      jamos = g_new (gunichar, max_jamos);
+	      memcpy (jamos, jamos_static, n_jamos*sizeof(gunichar));
 	    }
 	  else
-	    {
-	      /* Check syllable boundaries. */
-	      if ((IS_T (jamos[n_jamos - 1]) && IS_L (wc)) ||
-		  (IS_V (jamos[n_jamos - 1]) && IS_L (wc)) ||
-		  (IS_T (jamos[n_jamos - 1]) && IS_V (wc)))
-		{
-		  /* Draw a syllable.  */
-		  (*render_func) (font, subfont, jamos, n_jamos,
-				  glyphs, &n_glyphs, cluster_offset);
-		  cluster_offset = next - text;
-		  /* Clear.  */
-		  n_jamos = 0;
-		}
-	      if (n_jamos == max_jamos)
-		{
-		  max_jamos++;
-		  if (jamos == jamos_static)
-		    {
-		      jamos = g_new (gunichar2, max_jamos);
-		      memcpy(jamos, jamos_static, n_jamos*sizeof(gunichar2));
-		    }
-		  else
-		    jamos = g_renew (gunichar2, jamos, max_jamos);
-		}
-	      jamos[n_jamos++] = wc;
-	    }
-	}
-      else
-	{
-	  g_warning ("Character not handled by Hangul shaper: %#04x", wc4);
-	  continue;
+	    jamos = g_renew (gunichar, jamos, max_jamos);
 	}
 
-      ptr = next;
+      if (IS_S (wc))
+	{
+	  jamos[n_jamos++] = L_FROM_S (wc);
+	  jamos[n_jamos++] = V_FROM_S (wc);
+	  if (S_HAS_T (wc))
+	    jamos[n_jamos++] = T_FROM_S (wc);
+	}
+      else if (IS_M (wc) && !n_jamos)
+	;			/* ignore M's which do not follow syllables  */
+      else
+	jamos[n_jamos++] = wc;
+      p = g_utf8_next_char (p);
     }
 
   /* Draw the remaining Jamos.  */ 
-  if (n_jamos > 0)
-    {
-      (*render_func) (font, subfont, jamos, n_jamos,
-		      glyphs, &n_glyphs, cluster_offset);
-      cluster_offset = next - text;
-      n_jamos = 0;
-    }
+  if (n_jamos)
+    (*render_func) (font, subfont, jamos, n_jamos,
+		    glyphs, &n_glyphs, start - text);
 
   if (jamos != jamos_static)
     g_free (jamos);
@@ -836,7 +839,7 @@ hangul_engine_get_coverage (PangoFont  *font,
 	  for (i=0; i<KSC5601_HANGUL; i++)
 	    pango_coverage_set (result, __ksc5601_hangul_to_ucs[i], PANGO_COVERAGE_EXACT);
 	}
-      else if (render_func == render_syllable_with_ksx1005)
+      else if (render_func == render_syllable_with_ksx1001johab)
 	{
 	  for (i = 0x1100; i <= 0x11ff; i++)
 	    pango_coverage_set (result, i, PANGO_COVERAGE_EXACT);
