@@ -24,9 +24,9 @@
 
 #include <string.h>
 #include <stdlib.h>
-#include <stdio.h>
 #include <math.h>
 #include <glib.h>
+#include <glib/gprintf.h>
 
 #include <freetype/freetype.h>
 
@@ -200,7 +200,6 @@ FT_Face
 pango_ft2_font_get_face (PangoFont      *font)
 {
   PangoFT2Font *ft2font = (PangoFT2Font *)font;
-  FT_Face face;
   FT_Error error;
   FcPattern *pattern;
   FcChar8 *filename;
@@ -243,28 +242,20 @@ pango_ft2_font_get_face (PangoFont      *font)
 	bail0:
 	  load_fallback_face (ft2font, filename);
 	}
-      ft2font->face->generic.data = 0;
-    }
 
-  g_assert (ft2font->face);
-  
-  if (!set_unicode_charmap (ft2font->face))
-    {
-      g_warning ("Unable to load unicode charmap from font file %s", filename);
+      g_assert (ft2font->face);
       
-      FT_Done_Face (ft2font->face);
-      ft2font->face = NULL;
+      if (!set_unicode_charmap (ft2font->face))
+	{
+	  g_warning ("Unable to load unicode charmap from font file %s", filename);
+	  
+	  FT_Done_Face (ft2font->face);
+	  ft2font->face = NULL;
+	  
+	  load_fallback_face (ft2font, filename);
+	}
 
-      load_fallback_face (ft2font, filename);
-    }
-
-  face = ft2font->face;
-
-  if (ft2font->size != GPOINTER_TO_UINT (face->generic.data))
-    {
-      face->generic.data = GUINT_TO_POINTER (ft2font->size);
-
-      error = FT_Set_Char_Size (face,
+      error = FT_Set_Char_Size (ft2font->face,
 				PANGO_PIXELS_26_6 (ft2font->size),
 				PANGO_PIXELS_26_6 (ft2font->size),
 				0, 0);
@@ -272,7 +263,7 @@ pango_ft2_font_get_face (PangoFont      *font)
 	g_warning ("Error in FT_Set_Char_Size: %d", error);
     }
   
-  return face;
+  return ft2font->face;
 }
 
 static GType
@@ -1088,7 +1079,7 @@ _pango_ft2_ft_strerror (FT_Error error)
     {
       static char default_msg[100];
 
-      sprintf (default_msg, "Unknown FreeType2 error %#x", error);
+      g_sprintf (default_msg, "Unknown FreeType2 error %#x", error);
       return default_msg;
     }
 }
